<?php

namespace InSegment\ApiCore\Models;

use InSegment\ApiCore\Traits\Instantiatible;
use InSegment\ApiCore\Services\Transactor;
use InSegment\ApiCore\Services\BufferedBuilder;

use Illuminate\Support\Facades\Schema;
use Illuminate\Database\Eloquent\Model;

abstract class SupplementaryTable extends Model
{
    use Instantiatible;
    
    /**
     * Original table name
     *
     * @var string
     */
    protected $defaultTable;
    
    /**
     * UID
     *
     * @var string
     */
    public $uid;
    
    /**
     * Boot method
     * Subscribe to instantiation event
     * 
     * @return null
     */
    public static function boot()
    {
        static::instantiated(function ($model) {
            $instantiator = $model->_instantiator['model'] ?? null;
            if (isset($instantiator) && $instantiator->uid !== $model->uid) {
                $model->uid = $instantiator->uid;
                $model->updateTable();
            }
        });
    }
    
    /**
     * Constructor
     * 
     * @param array $attributes
     * @param string|null $uid
     */
    public function __construct(array $attributes = array(), $uid = null)
    {
        parent::__construct($attributes);
        $this->defaultTable = $this->getTable();
        $this->uid = $uid ?? Transactor::getTransactionUID();
        $this->updateTable();
    }
    
    /**
     * Update table of the model
     * 
     * @return $this
     */
    protected function updateTable()
    {
        $database = Transactor::getTransactionSchema();
        $table = $this->tableNameGenerator($this->uid);
        $this->table = "{$database}.{$table}";
        return $this;
    }

    /**
     * Table name generator
     * 
     * @return string
     */
    protected function tableNameGenerator()
    {
        return isset($this->uid) ? substr(implode('_', [$this->defaultTable, $this->uid]), 0, 64) : '';
    }
    
    /**
     * Indicate there is no incrementing key
     * 
     * @var bool 
     */
    public $incrementing = false;

    /**
     * The model has timestamps
     * 
     * @var bool
     */
    public $timestamps = false;
    
    /**
     * Create a new Eloquent query builder for the model.
     *
     * @param  \Illuminate\Database\Query\Builder  $query
     * @return \InSegment\ApiCore\Services\BufferedBuilder|static
     */
    public function newEloquentBuilder($query)
    {
        return new BufferedBuilder($query);
    }
    
    /**
     * Run the migrations.
     *
     * @param array $options
     * @param bool $inMemory
     * @param string|null $uid
     * @return void
     */
    public abstract static function establish(array $options, bool $inMemory, $uid = null);

    /**
     * Reverse the migrations.
     *
     * @param string|null $uid
     * @return void
     */
    public static function disband($uid = null)
    {
        Schema::dropIfExists((new static([], $uid))->getTable());
    }
    
    /**
     * Get query optionally specifying $uid
     * 
     * @param string|null $uid
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public static function query($uid = null)
    {
        return (new static([], $uid))->newQuery();
    }
}
